#include "dffimp.h"

HINSTANCE hInstance;

TCHAR *GetString(int id)
{
	static TCHAR buf[256];
	if (hInstance)
		return LoadString(hInstance, id, buf, sizeof(buf)) ? buf : NULL;
	return NULL;
}

void onStartup(void *param, NotifyInfo *info) { initImporter(); }

BOOL WINAPI DllMain(HINSTANCE hinstDLL,ULONG fdwReason,LPVOID lpvReserved)
{
	hInstance = hinstDLL;
	switch(fdwReason){
	case DLL_PROCESS_ATTACH:
		RegisterNotification(onStartup, NULL, NOTIFY_SYSTEM_STARTUP);
		break;
	case DLL_THREAD_ATTACH:
		break;
	case DLL_THREAD_DETACH:
		break;
	case DLL_PROCESS_DETACH:
		break;
	}
	return TRUE;
}


//------------------------------------------------------

class DFFImportDesc:public ClassDesc {
	public:
	int IsPublic() { return 1; }
	void *Create(BOOL loading = FALSE) { return new DFFImport; }
	const TCHAR *ClassName() { return GetString(IDS_CLASSNAME); }
	SClass_ID SuperClassID() { return SCENE_IMPORT_CLASS_ID; }
	Class_ID ClassID() { return Class_ID(0xc42486c, 0x6f900bef); }
	const TCHAR *Category() { return GetString(IDS_SCENEIMPORT);  }
};

static DFFImportDesc importDesc;

__declspec(dllexport) const TCHAR*
LibDescription(void) { return GetString(IDS_LIBDESCRIPTION); }

__declspec(dllexport) int
LibNumberClasses(void) { return 1; }

__declspec( dllexport ) ClassDesc*
LibClassDesc(int i)
{
	switch(i){
	case 0:
		return &importDesc;
	default:
		return NULL;
	}
}

// Return version so can detect obsolete DLLs
__declspec(dllexport) ULONG 
LibVersion(void) { return VERSION_3DSMAX; }

// Let the plug-in register itself for deferred loading
__declspec(dllexport)
ULONG CanAutoDefer(void)
{
	return 1;
}


//
// DFFImport
//

DFFImport::DFFImport(void)
{
}

DFFImport::~DFFImport(void)
{
}

int
DFFImport::ExtCount(void)
{
	return 1;
}

const TCHAR*
DFFImport::Ext(int n)
{
	switch(n){
	case 0:
		return _T("DFF");
	}
	return _T("");
}

const TCHAR*
DFFImport::LongDesc(void)
{
	return GetString(IDS_DFFFILE);
}
	
const TCHAR*
DFFImport::ShortDesc(void)
{
	return GetString(IDS_CLASSNAME);
}

const TCHAR*
DFFImport::AuthorName(void)
{
	return GetString(IDS_AUTHOR);
}

const TCHAR*
DFFImport::CopyrightMessage(void)
{
	return GetString(IDS_COPYRIGHT);
}

const TCHAR *
DFFImport::OtherMessage1(void)
{
	return _T("");
}

const TCHAR *
DFFImport::OtherMessage2(void)
{
	return _T("");	
}

unsigned int
DFFImport::Version(void)
{				// Version number * 100 (i.e. v3.01 = 301)
	return 100;
}

void
DFFImport::ShowAbout(HWND hWnd)
{
}

int
findFirstVertex(float *verts, int n, float *v)
{
	for(int i = 0; i < n; i++) {
		if(verts[0] == v[0] && verts[1] == v[1] && verts[2] == v[2])
			return i;
		verts += 3;
	}
	/* shouldn't happen */
	return -1;
}

int
findFirstTexCoord(float *verts, int n, float *v)
{
	for(int i = 0; i < n; i++) {
		if(verts[0] == v[0] && verts[1] == v[1])
			return i;
		verts += 2;
	}
	/* shouldn't happen */
	return -1;
}

int
findFirstColor(rw::uint8 *verts, int n, rw::uint8 *v)
{
	for(int i = 0; i < n; i++) {
		if(verts[0] == v[0] && verts[1] == v[1] &&
		   verts[2] == v[2] && verts[3] == v[3])
			return i;
		verts += 4;
	}
	/* shouldn't happen */
	return -1;
}

void
animateTexture(rw::Animation *anim, BitmapTex *tex)
{
	using namespace rw;
	if(tex == NULL)
		return;
	UVAnimCustomData *cust = (UVAnimCustomData*)anim->customData;
	StdUVGen *uvgen = tex->GetUVGen();
	int channel = uvgen->GetMapChannel();
	if(cust->nodeToUVChannel[0] != channel-1)
		return;
	if(anim->interpInfo->id != 0x1C1){
		lprintf(_T("warning: cannot import animations of type %x\n"),
		        anim->interpInfo->id);
		return;
	}
	UVAnimKeyFrame *kf = (UVAnimKeyFrame*)anim->keyframes;
	AnimateOn();
	for(int32 i = 0; i < anim->numFrames; i++){
		TimeValue t = SecToTicks(kf->time);
		uvgen->SetWAng(kf->uv[0], t);
		uvgen->SetUScl(kf->uv[1], t);
		uvgen->SetVScl(kf->uv[2], t);
		// no skew :/
		uvgen->SetUOffs(-kf->uv[4], t);
		uvgen->SetVOffs(kf->uv[5], t);
		kf++;
	}
	AnimateOff();
}

BitmapTex*
makeTexture(rw::Texture *tex)
{
	using namespace rw;
	static WCHAR fuckyou[MAX_PATH];
	static char texname[MAX_PATH];

	BitmapTex *bmtex;
	if(tex == NULL)
		return NULL;
	bmtex = NewDefaultBitmapTex();
	strncpy(texname, tex->name, 32);
	strcat(texname, ".tga");
#ifdef _UNICODE
	mbstowcs(fuckyou, texname, MAX_PATH);
	bmtex->SetMapName(fuckyou);
#else
	bmtex->SetMapName(texname);
#endif
	switch(tex->filterAddressing & 0xFF){
	case Texture::NEAREST:
		bmtex->SetFilterType(FILTER_NADA);
		break;
	case Texture::LINEAR:
		bmtex->SetFilterType(FILTER_SAT);
		break;
	case Texture::LINEARMIPLINEAR:
		bmtex->SetFilterType(FILTER_PYR);
		break;
	}
	switch(tex->filterAddressing>>8 & 0xF){
	case Texture::WRAP:
		bmtex->GetUVGen()->SetFlag(U_WRAP, 1);
		break;
	case Texture::MIRROR:
		bmtex->GetUVGen()->SetFlag(U_MIRROR, 1);
		break;
	}
	switch(tex->filterAddressing>>12 & 0xF){
	case Texture::WRAP:
		bmtex->GetUVGen()->SetFlag(V_WRAP, 1);
		break;
	case Texture::MIRROR:
		bmtex->GetUVGen()->SetFlag(V_MIRROR, 1);
		break;
	}
	return bmtex;
}

void
DFFImport::makeMaterials(rw::Atomic *a, INode *inode)
{
	using namespace rw;
	using namespace gta;

	Geometry *g = a->geometry;
	MultiMtl *multi = NULL;

	if(g->numMaterials == 0)
		return;	// can this happen?
	if(g->numMaterials > 1){
		multi = NewDefaultMultiMtl();
		multi->SetNumSubMtls(g->numMaterials);
	}

	for(int32 i = 0; i < g->numMaterials; i++){
		rw::Material *m = g->materialList[i];
		Mtl *mat = (Mtl*)ifc->CreateInstance(SClass_ID(MATERIAL_CLASS_ID), Class_ID(0x29b71842, 0x52508b70));
		IParamBlock2 *pb = mat->GetParamBlock(0);

		pb->SetValue(mat_color, 0, Color(m->color.red/255.0f, m->color.green/255.0f, m->color.blue/255.0f));
		pb->SetValue(mat_coloralpha, 0, m->color.alpha/255.0f);
		pb->SetValue(mat_sp_ambient, 0, m->surfaceProps.ambient);
		pb->SetValue(mat_sp_diffuse, 0, m->surfaceProps.diffuse);
		pb->SetValue(mat_sp_specular, 0, m->surfaceProps.specular);

		BitmapTex *diffmap = NULL;
		if(m->texture)
			mat->GetParamBlock(0)->SetValue(mat_texmap_texture, 0, diffmap = makeTexture(m->texture));

		MatFX *matfx = *PLUGINOFFSET(MatFX*, m, matFXGlobals.materialOffset);
		int32 idx;
		BitmapTex *bmtex;
		BitmapTex *dualmap = NULL;
		if(matfx){
			pb->SetValue(mat_matfxeffect, 0, (int)matfx->type+1);
			switch(matfx->type){
			case MatFX::BUMPMAP:
				idx = matfx->getEffectIndex(MatFX::BUMPMAP);
				assert(idx >= 0);
				bmtex = makeTexture(matfx->fx[idx].bump.bumpedTex);
				pb->SetValue(mat_texmap_bumpmap, 0, bmtex);
				pb->SetValue(mat_bumpmap_amount, 0, matfx->fx[idx].bump.coefficient);
				break;
	
			case MatFX::ENVMAP:
				idx = matfx->getEffectIndex(MatFX::ENVMAP);
				assert(idx >= 0);
				bmtex = makeTexture(matfx->fx[idx].env.tex);
				pb->SetValue(mat_texmap_envmap, 0, bmtex);
				pb->SetValue(mat_envmap_amount, 0, matfx->fx[idx].env.coefficient);
				break;
	
			case MatFX::DUAL:
				idx = matfx->getEffectIndex(MatFX::DUAL);
				assert(idx >= 0);
				dualmap = makeTexture(matfx->fx[idx].dual.tex);
				if(dualmap)
					dualmap->GetUVGen()->SetMapChannel(2);
				pb->SetValue(mat_texmap_pass2, 0, dualmap);
				pb->SetValue(mat_pass2_srcblend, 0, matfx->fx[idx].dual.srcBlend);
				pb->SetValue(mat_pass2_destblend, 0, matfx->fx[idx].dual.dstBlend);
				break;
	
			case MatFX::BUMPENVMAP:
				idx = matfx->getEffectIndex(MatFX::ENVMAP);
				assert(idx >= 0);
				bmtex = makeTexture(matfx->fx[idx].env.tex);
				pb->SetValue(mat_texmap_envmap, 0, bmtex);
				pb->SetValue(mat_envmap_amount, 0, matfx->fx[idx].env.coefficient);
				idx = matfx->getEffectIndex(MatFX::BUMPMAP);
				assert(idx >= 0);
				bmtex = makeTexture(matfx->fx[idx].bump.bumpedTex);
				pb->SetValue(mat_texmap_bumpmap, 0, bmtex);
				pb->SetValue(mat_bumpmap_amount, 0, matfx->fx[idx].bump.coefficient);
				break;
			}
		}

		// R* extensions
		IParamBlock2 *pbr = mat->GetParamBlock(1);
		EnvMat *env = *PLUGINOFFSET(EnvMat*, m, envMatOffset);
		if(env){
			pbr->SetValue(mat_enEnv, 0, TRUE, 0);
			pbr->SetValue(mat_scaleX, 0, env->scaleX/8.0f);
			pbr->SetValue(mat_scaleY, 0, env->scaleY/8.0f);
			pbr->SetValue(mat_transScaleX, 0, env->transScaleX/8.0f);
			pbr->SetValue(mat_transScaleY, 0, env->transScaleY/8.0f);
			pbr->SetValue(mat_shininess, 0, env->shininess/8.0f);
		}

		SpecMat *spec = *PLUGINOFFSET(SpecMat*, m, specMatOffset);
		if(spec){
			pbr->SetValue(mat_enSpec, 0, TRUE, 0);
			pbr->SetValue(mat_specularity, 0, spec->specularity);
			pbr->SetValue(mat_specMap, 0, makeTexture(spec->texture));
		}

		UVAnim *uvanim = PLUGINOFFSET(UVAnim, m, uvAnimOffset);
		for(int j = 0; j < 8; j++)
			if(uvanim->interp[i]){
				animateTexture(uvanim->interp[j]->anim, diffmap);
				animateTexture(uvanim->interp[j]->anim, dualmap);
			}

		if(g->numMaterials == 1){
			inode->SetMtl(mat);
			return;
		}
		MSTR str = _T("");
		multi->SetSubMtlAndName(i, mat, str);
	}
	inode->SetMtl(multi);
}

void
DFFImport::makeMesh(rw::Atomic *a, Mesh *maxmesh)
{
	using namespace rw;

	Geometry *g = a->geometry;

	float *gverts = g->morphTargets[0].vertices;
	int *newind = new int[g->numVertices];
	for(int32 i = 0; i < g->numVertices; i++){
		int j = findFirstVertex(gverts, g->numVertices, &gverts[i*3]);
		assert(j >= 0);
		newind[i] = j;
	}
	maxmesh->setNumVerts(g->numVertices);
	Point3 maxvert;
	for(int32 i = 0; i < g->numVertices; i++){
		int j = newind[i];
		maxvert[0] = gverts[j*3+0];
		maxvert[1] = gverts[j*3+1];
		maxvert[2] = gverts[j*3+2];
		maxmesh->setVert(i, maxvert);
	}
	maxmesh->setNumFaces(g->numTriangles);
	for(int32 i = 0; i < g->numTriangles; i++){
		maxmesh->faces[i].setVerts(newind[g->triangles[i].v[0]],
		                           newind[g->triangles[i].v[1]],
		                           newind[g->triangles[i].v[2]]);
		maxmesh->faces[i].setSmGroup(0);
		maxmesh->faces[i].setMatID(g->triangles[i].matId);
		maxmesh->faces[i].setEdgeVisFlags(1, 1, 1);
	}

	for(int k = 0; k < g->numTexCoordSets; k++){
		maxmesh->setMapSupport(MAP_TEXCOORD0+k, TRUE);
		float *uvs = g->texCoords[k];
		for(int32 i = 0; i < g->numVertices; i++){
			int j = findFirstTexCoord(uvs, g->numVertices, &uvs[i*2]);
			assert(j >= 0);
			newind[i] = j;
		}
		maxmesh->setNumMapVerts(MAP_TEXCOORD0+k, g->numVertices);
		Point3 maxvert;
		for(int32 i = 0; i < g->numVertices; i++){
			int j = newind[i];
			maxvert[0] = uvs[j*2+0];
			maxvert[1] = 1-uvs[j*2+1];
			maxvert[2] = 0.0f;
			maxmesh->setMapVert(MAP_TEXCOORD0+k, i, maxvert);
		}
		maxmesh->setNumMapFaces(MAP_TEXCOORD0+k, g->numTriangles);
		TVFace *tvf = maxmesh->mapFaces(MAP_TEXCOORD0+k);
		for(int32 i = 0; i < g->numTriangles; i++){
			tvf[i].setTVerts(newind[g->triangles[i].v[0]],
			                 newind[g->triangles[i].v[1]],
			                 newind[g->triangles[i].v[2]]);
		}
	}

	if(g->geoflags & rw::Geometry::PRELIT){
		if(!maxmesh->mapSupport(MAP_ALPHA))
			maxmesh->setMapSupport(MAP_ALPHA, TRUE);
		uint8 *prelit = g->colors;
		for(int32 i = 0; i < g->numVertices; i++){
			int j = findFirstColor(prelit, g->numVertices, &prelit[i*4]);
			assert(j >= 0);
			newind[i] = j;
		}
		maxmesh->setNumVertCol(g->numVertices);
		maxmesh->setNumMapVerts(MAP_ALPHA, g->numVertices);
		Point3 maxvert;
		for(int32 i = 0; i < g->numVertices; i++){
			int j = newind[i];
			maxvert[0] = prelit[j*4+0]/255.0f;
			maxvert[1] = prelit[j*4+1]/255.0f;
			maxvert[2] = prelit[j*4+2]/255.0f;
			maxmesh->vertCol[i] = maxvert;
			maxvert[0] = maxvert[1] = maxvert[2] = prelit[j*4+3]/255.0f;
			maxmesh->setMapVert(MAP_ALPHA, i, maxvert);
		}
		maxmesh->setNumVCFaces(g->numTriangles);
		maxmesh->setNumMapFaces(MAP_ALPHA, g->numTriangles);
		TVFace *af = maxmesh->mapFaces(MAP_ALPHA);
		for(int32 i = 0; i < g->numTriangles; i++){
			maxmesh->vcFace[i].setTVerts(newind[g->triangles[i].v[0]],
			                             newind[g->triangles[i].v[1]],
			                             newind[g->triangles[i].v[2]]);
			af[i].setTVerts(maxmesh->vcFace[i].getAllTVerts());
		}
	}

	gta::ExtraVertColors *colordata = PLUGINOFFSET(gta::ExtraVertColors, g, gta::extraVertColorOffset);
	if(colordata->nightColors){
		maxmesh->setMapSupport(MAP_EXTRACOLORS, TRUE);
		maxmesh->setMapSupport(MAP_EXTRAALPHA, TRUE);
		uint8 *prelit = colordata->nightColors;
		for(int32 i = 0; i < g->numVertices; i++){
			int j = findFirstColor(prelit, g->numVertices, &prelit[i*4]);
			assert(j >= 0);
			newind[i] = j;
		}
		maxmesh->setNumMapVerts(MAP_EXTRACOLORS, g->numVertices);
		maxmesh->setNumMapVerts(MAP_EXTRAALPHA, g->numVertices);
		Point3 maxvert;
		for(int32 i = 0; i < g->numVertices; i++){
			int j = newind[i];
			maxvert[0] = prelit[j*4+0]/255.0f;
			maxvert[1] = prelit[j*4+1]/255.0f;
			maxvert[2] = prelit[j*4+2]/255.0f;
			maxmesh->setMapVert(MAP_EXTRACOLORS, i, maxvert);
			maxvert[0] = maxvert[1] = maxvert[2] = prelit[j*4+3]/255.0f;
			maxmesh->setMapVert(MAP_EXTRAALPHA, i, maxvert);
		}
		maxmesh->setNumMapFaces(MAP_EXTRACOLORS, g->numTriangles);
		maxmesh->setNumMapFaces(MAP_EXTRAALPHA, g->numTriangles);
		TVFace *cf = maxmesh->mapFaces(MAP_EXTRACOLORS);
		TVFace *af = maxmesh->mapFaces(MAP_EXTRAALPHA);
		for(int32 i = 0; i < g->numTriangles; i++){
			cf[i].setTVerts(newind[g->triangles[i].v[0]],
			                newind[g->triangles[i].v[1]],
			                newind[g->triangles[i].v[2]]);
			af[i].setTVerts(cf[i].getAllTVerts());
		}
	}

	maxmesh->DeleteIsoVerts();
	maxmesh->DeleteIsoMapVerts();

	// cargo cult
	maxmesh->buildNormals();
	maxmesh->buildBoundingBox();
	maxmesh->InvalidateEdgeList();
	maxmesh->InvalidateGeomCache();

	if(DFFImport::autoSmooth)
		maxmesh->AutoSmooth(DegToRad(45.0f), 0, 1);

	delete[] newind;
}

void
getBonePositions(rw::Skin *skin, rw::HAnimHierarchy *hier, Point3 *positions)
{
	using namespace rw;

	Matrix3 matrix;
	for(int32 i = 0; i < skin->numBones; i++){
		float *m = &skin->inverseMatrices[i*16];
		matrix.SetRow(0, Point3(m[0],  m[1],  m[2]));
		matrix.SetRow(1, Point3(m[4],  m[5],  m[6]));
		matrix.SetRow(2, Point3(m[8],  m[9],  m[10]));
		matrix.SetRow(3, Point3(m[12], m[13], m[14]));
		matrix.Invert();
		Point3 pos = matrix.GetRow(3);
		*positions++ = pos;
	}
}

//rw::Atomic*
//getSkinAtomic(rw::Clump *c)
//{
//	using namespace rw;
//
//	FORLIST(lnk, c->atomics){
//		Atomic *a = Atomic::fromClump(lnk);
//		Skin *s = *PLUGINOFFSET(Skin*, a->geometry, skinGlobals.offset);
//		if(s)
//			return a;
//	}
//	return NULL;
//}

int
findIdx(rw::HAnimHierarchy *hier, Point3 pos, Point3 *positions)
{
	float dist = 10000.0f;
	int i = 0;
	for(int j = 0; j < hier->numNodes; j++){
		float d = (pos-positions[j]).Length();
		if(d < dist){
			dist = d;
			i = j;
		}
	}
	if(dist < 0.05)
		return i;
	return -1;
}

int
findVertex(rw::Geometry *g, Point3 *v)
{
	using namespace rw;
	float *verts = g->morphTargets[0].vertices;
	for(int32 i = 0; i < g->numVertices; i++){
		if(abs(verts[i*3+0]-v->x) < 0.0001f &&
		   abs(verts[i*3+1]-v->y) < 0.0001f &&
		   abs(verts[i*3+2]-v->z) < 0.0001f)
			return i;
	}
	return -1;
}

::Mesh*
getMesh(INode *node, TimeValue t)
{
	::Object *o = node->EvalWorldState(t).obj;
	if(o->CanConvertToType(Class_ID(TRIOBJ_CLASS_ID, 0))){
		TriObject *tri = (TriObject*)o->ConvertToType(t, Class_ID(TRIOBJ_CLASS_ID, 0));
		// TODO: if tri != o delete!
		return &tri->GetMesh();
	}
	return NULL;
}

rw::Skin*
getSkin(rw::Atomic *a)
{
	return *PLUGINOFFSET(rw::Skin*, a->geometry, rw::skinGlobals.offset);
}

BOOL
DFFImport::dffFileRead(const TCHAR *filename)
{
	using namespace rw;

	StreamFile in;
	Clump *c;

#ifdef _UNICODE
	char path[MAX_PATH];
	wcstombs(path, filename, MAX_PATH);
	if(in.open(path, "rb") == NULL)
#else
	if(in.open(filename, "rb") == NULL)
#endif
		return 0;
	currentUVAnimDictionary = NULL;
	currentTexDictionary = TexDictionary::create();
	debugFile = (char*)filename;
	ChunkHeaderInfo header;
	readChunkHeaderInfo(&in, &header);
	if(header.type == ID_UVANIMDICT){
		UVAnimDictionary *dict = UVAnimDictionary::streamRead(&in);
		currentUVAnimDictionary = dict;
		readChunkHeaderInfo(&in, &header);
	}
	if(header.type != ID_CLUMP){
		in.close();
		return 0;
	}
	c = Clump::streamRead(&in);
	in.close();
	if(c == NULL)
		return 0;

	int32 platform = findPlatform(c);
	if(platform){
		rw::platform = platform;
		switchPipes(c, platform);
	}

	FORLIST(lnk, c->atomics){
		Atomic *a = Atomic::fromClump(lnk);
		ObjPipeline *p = a->getPipeline();
		p->uninstance(a);
	}

	TimeValue t = ifc->GetTime();
	int32 numFrames = c->getFrame()->count();
	Frame **flist = new Frame*[numFrames];
	makeFrameList(c->getFrame(), flist);

	HAnimHierarchy *hier = HAnimHierarchy::find(c->getFrame());
	if(hier)
		hier->attach();
	//Atomic *skinAtm = getSkinAtomic(c);
	//Skin *rwskin = NULL;
	//if(skinAtm)
	//	rwskin = *PLUGINOFFSET(Skin*, skinAtm->geometry, skinGlobals.offset);
	//INode *skinNode = NULL;
	INode **skinNodes = new INode*[c->countAtomics()];
	Atomic **skinAtomics = new Atomic*[c->countAtomics()];
	int numSkinned = 0;

	INode *bones[256];
	memset(bones, 0, sizeof(bones));

	INode **nodelist = new INode*[numFrames];
	for(int j = 0; j < numFrames; j++){
		flist[j]->updateLTM();
		Frame *f = flist[j];
		//Point3 pos(f->ltm[12], f->ltm[13], f->ltm[14]);

		INode *node;
		// Dummy as placeholder
		node = ifc->CreateObjectNode((::Object*)ifc->CreateInstance(HELPER_CLASS_ID, Class_ID(DUMMY_CLASS_ID, 0)));
		nodelist[j] = node;
		Matrix3 tm;
		MRow *m = tm.GetAddr();
		f->updateLTM();
		float *rwm = f->ltm;
		m[0][0] = rwm[0];
		m[0][1] = rwm[1];
		m[0][2] = rwm[2];
		m[1][0] = rwm[4];
		m[1][1] = rwm[5];
		m[1][2] = rwm[6];
		m[2][0] = rwm[8];
		m[2][1] = rwm[9];
		m[2][2] = rwm[10];
		m[3][0] = rwm[12];
		m[3][1] = rwm[13];
		m[3][2] = rwm[14];
		node->SetNodeTM(t, tm);
		int isbone = 0;

		// HAnim
		if(hier){
			HAnimData *hanim = HAnimData::get(f);
			if(hanim->id >= 0){
				isbone = 1;
				node->SetUserPropInt(_T("tag"), hanim->id);
				bones[hier->getIndex(hanim->id)] = node;
			}
		}

		// GTA Node name
		char *name = gta::getNodeName(f);
		char maxname[32];
		if(name[0]){
			if(DFFImport::prepend){
				maxname[0] = '!';
				maxname[1] = '\0';
				strcat(maxname, name);
			}else
				strncpy(maxname, name, 32);
#ifdef _UNICODE
			WCHAR fuckyou[50];
			mbstowcs(fuckyou, maxname, 50);
			node->SetName(fuckyou);
#else
			node->SetName(maxname);
#endif
		}

		// Objects
		int flipz = 0;
		FORLIST(lnk, f->objectList){
			ObjectWithFrame *obj = ObjectWithFrame::fromFrame(lnk);
			if(obj->type == Atomic::ID){
				Atomic *a = (Atomic*)obj;
				TriObject *tri = CreateNewTriObject();
				if(!tri) return 0;
				::Mesh *msh = &tri->GetMesh();
				makeMesh(a, msh);
				// make a new node outside the hierarchy for skinned objects
				if(getSkin(a)){
					INode *skinNode = ifc->CreateObjectNode(tri);
					skinNode->SetNodeTM(t, node->GetNodeTM(t));
					makeMaterials(a, skinNode);
					skinAtomics[numSkinned] = a;
					skinNodes[numSkinned] = skinNode;
					numSkinned++;
				}else{
					node->SetObjectRef(tri);
					makeMaterials(a, node);
					break;
				}
			}else if(obj->type == rw::Light::ID){
				rw::Light *l = (rw::Light*)obj;
				int maxtype;
				int type = l->getType();
				switch(type){
				case rw::Light::SOFTSPOT:
				case rw::Light::SPOT:
					maxtype = FSPOT_LIGHT;
					break;
				case rw::Light::DIRECTIONAL:
					maxtype = DIR_LIGHT;
					break;
				case rw::Light::POINT:
				case rw::Light::AMBIENT:
				default:
					maxtype = OMNI_LIGHT;
					break;
				}
				GenLight *gl = impifc->CreateLightObject(maxtype);
				gl->SetRGBColor(t, Point3(l->color.red, l->color.green, l->color.blue));
				if(type >= rw::Light::POINT)
					gl->SetAtten(t, ATTEN_END, l->radius);
				if(maxtype == FSPOT_LIGHT){
					float32 a = l->getAngle()*2*180.0f/pi;
					gl->SetFallsize(t, a);
					a -= type == rw::Light::SPOT ? 2.0f : 10.0f;
					gl->SetHotspot(t, a);
				}
				gl->SetAmbientOnly(type == rw::Light::AMBIENT);
				node->SetObjectRef(gl);
				node->SetWireColor(0xFF00E5FF);
				ifc->AddLightToScene(node);
				gl->Enable(TRUE);
				flipz = 1;
				break;
			}else if(obj->type == rw::Camera::ID){
				rw::Camera *c = (rw::Camera*)obj;
				GenCamera *gc = impifc->CreateCameraObject(c->projection == 1 ? FREE_CAMERA : PARALLEL_CAMERA);
				node->SetObjectRef(gc);
				gc->SetFOV(t, atan(c->viewWindow.x)*2);
				gc->SetManualClip(1);
				gc->SetClipDist(t, CAM_HITHER_CLIP, c->nearPlane);
				gc->SetClipDist(t, CAM_YON_CLIP, c->farPlane);
				gc->Enable(TRUE);
				flipz = 1;
				break;
			}
		}

		if(flipz){
			AngAxis aa(Point3(0,1,0), pi);
			node->Rotate(t, tm, aa);
		}
	}

	/* build hierarchy */
	INode *rootnode = NULL;
	for(int j = 0; j < numFrames; j++){
		if(flist[j]->getParent() == NULL){
			rootnode = nodelist[j];
			continue;
		}
		int frm = findPointer(flist[j]->getParent(), (void**)flist, numFrames);
		nodelist[frm]->AttachChild(nodelist[j], 1);
	}

	// Skin
	for(int i = 0; i < numSkinned; i++){
		INode *skinNode = skinNodes[i];
		Skin *rwskin = getSkin(skinAtomics[i]);
		if(rwskin->numBones != hier->numNodes){
			lprintf(_T("number of bones doesn't match hierarchy\n"));
			goto noskin;
		}
		for(int j = 0; j < hier->numNodes; j++){
			//lprintf("%d %p\n", j, bones[j]);
			if(bones[j] == NULL){
				lprintf(_T("bone missing\n"));
				goto noskin;
			}
			bones[j]->ShowBone(2);
			bones[j]->SetWireColor(0xFF00FFFF);
		}
		Modifier *skin = (Modifier*)ifc->CreateInstance(SClass_ID(OSM_CLASS_ID), SKIN_CLASSID);
		ISkinImportData *skinImp = (ISkinImportData*)skin->GetInterface(I_SKINIMPORTDATA);
		GetCOREInterface7()->AddModifier(*skinNode, *skin);
		for(int j = 0; j < hier->numNodes; j++)
			skinImp->AddBoneEx(bones[j], true);
		// Assigning weights becomes a bit complicated because
		// we deleted isolated vertices.
		Tab<::INode*> bt;
		Tab<float> wt;
		bt.SetCount(4, 1);
		wt.SetCount(4, 1);
		skinNode->EvalWorldState(t);
		float *w;
		uint8 *ix;
		::Mesh *maxmesh = getMesh(skinNode, t);
		assert(maxmesh != NULL);
		int numVerts = maxmesh->getNumVerts();
		for(int j = 0; j < numVerts; j++){
			Point3 pos = maxmesh->getVert(j);
			int idx = findVertex(skinAtomics[i]->geometry, &pos);
			assert(idx >= 0);
			w = &rwskin->weights[idx*4];
			ix = &rwskin->indices[idx*4];
			for(int k = 0; k < 4; k++){
				bt[k] = bones[ix[k]];
				wt[k] = w[k];
			}
			skinImp->AddWeights(skinNode, j, bt, wt);
		}
	}
	noskin:

	if(DFFImport::convertAxes){
		AngAxis aa(Point3(1,0,0), -pi/2);
		Matrix3 tm;
		tm.IdentityMatrix();
		rootnode->Rotate(t, tm, aa);
	}

	impifc->RedrawViews();
	delete[] skinAtomics;
	delete[] skinNodes;
	delete[] nodelist;
	delete[] flist;

	if(currentUVAnimDictionary)
		currentUVAnimDictionary->destroy();
	currentTexDictionary->destroy();
	c->destroy();

	rw::platform = rw::PLATFORM_D3D8;
	return TRUE;
}

static INT_PTR CALLBACK ImportDlgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	DFFImport *imp = DLGetWindowLongPtr<DFFImport*>(hWnd); 
	switch(msg){
	case WM_INITDIALOG:
		imp = (DFFImport*)lParam;
		DLSetWindowLongPtr(hWnd, lParam); 
		CenterWindow(hWnd, GetParent(hWnd)); 
		CheckDlgButton(hWnd, IDC_CONVAXIS, DFFImport::convertAxes);
		CheckDlgButton(hWnd, IDC_AUTOSMOOTH, DFFImport::autoSmooth);
		break;
	case WM_COMMAND:
		switch(LOWORD(wParam)){
		case IDOK:
			DFFImport::convertAxes = IsDlgButtonChecked(hWnd, IDC_CONVAXIS); 
			DFFImport::autoSmooth = IsDlgButtonChecked(hWnd, IDC_AUTOSMOOTH); 
			EndDialog(hWnd, 1);
			break;
		case IDCANCEL:
			EndDialog(hWnd, 0);
			break;
		default:
			//lprintf("WM_COMMAND: %d\n", LOWORD(wParam));
			break;
		}
		break;
		default:
			//lprintf("MESG: %d\n", LOWORD(msg));
			return FALSE;
	}
	return TRUE;
}

int DFFImport::convertAxes;
int DFFImport::autoSmooth = 1;
int DFFImport::prepend;

int rwInitialized;

void
initRW(void)
{
	rw::version = 0x33002;
	//rw::version = 0x35000;
	rw::platform = rw::PLATFORM_D3D8;
	//rw::platform = rw::PLATFORM_PS2;
	gta::attachPlugins();
	rwInitialized = 1;
}

int
DFFImport::DoImport(const TCHAR *filename, ImpInterface *ii, Interface *i, BOOL suppressPrompts)
{
	if(!rwInitialized)
		initRW();
	this->ifc = i;
	this->impifc = ii;

	if(!suppressPrompts){
		DFFImport::prepend = 0;
		if (!DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_DFFIMPORT),
			i->GetMAXHWnd(), ImportDlgProc, (LPARAM)this)) {
			return 1;
		}
	}

	if(dffFileRead(filename))
		return 1;
	return 0;
}
