#include "dffimp.h"

Value *getConvertAxes() { return Integer::intern(DFFImport::convertAxes); }
Value *setConvertAxes(Value *val) { DFFImport::convertAxes = val->to_int(); return val; }

Value *getAutoSmooth() { return Integer::intern(DFFImport::autoSmooth); }
Value *setAutoSmooth(Value *val) { DFFImport::autoSmooth = val->to_int(); return val; }

Value *getPrepend() { return Integer::intern(DFFImport::prepend); }
Value *setPrepend(Value *val) { DFFImport::prepend = val->to_int(); return val; }

static float flipZMat[16] = {
	-1.0f, 0.0f,  0.0f, 0.0f,
	 0.0f, 1.0f,  0.0f, 0.0f,
	 0.0f, 0.0f, -1.0f, 0.0f,
	 0.0f, 0.0f,  0.0f, 0.0f
};

void
dumpScene(INode *node, int ind)
{
	for(int i = 0; i < ind; i++)
		lprintf(_T(" "));
	lprintf(_T("%s\n"), node->GetName());
	int numChildren = node->NumberOfChildren();
	for(int i = 0; i < numChildren; i++)
		dumpScene(node->GetChildNode(i), ind+1);
}

INode*
getRootOf(INode *node)
{
	while(!node->GetParentNode()->IsRootNode())
		node = node->GetParentNode();
	return node;
}

void
dumpFrameHier(rw::Frame *frame, int ind = 0)
{
	using namespace rw;
	for(int i = 0; i < ind; i++)
		printf(" ");
	char *name = gta::getNodeName(frame);
	HAnimData *hanim = PLUGINOFFSET(HAnimData, frame, hAnimOffset);
	printf("%s %d %d %s\n", name[0] ? name : "---", frame->objectList.count(), hanim->id, hanim->hierarchy ? "HIERARCHY" : "");
	for(Frame *child = frame->child;
	    child; child = child->next)
		dumpFrameHier(child, ind+1);
}

void
convertLight(rw::Frame *frame, rw::Clump *clump, INode *node)
{
	using namespace rw;
	GenLight *gl = (GenLight*)node->EvalWorldState(0).obj;
	LightState state;
	Interval valid = FOREVER;
	gl->EvalLightState(0, valid, &state);
	int32 rwtype = rw::Light::POINT;
	switch(state.type){
	case OMNI_LGT:
		lprintf(_T("point\n"));
		rwtype = rw::Light::POINT;
		break;
	case SPOT_LGT:
		lprintf(_T("spot\n"));
		rwtype = rw::Light::SPOT;
		break;
	case DIRECT_LGT:
		lprintf(_T("direct\n"));
		rwtype = rw::Light::DIRECTIONAL;
		break;
	case AMBIENT_LGT:
		lprintf(_T("ambient\n"));
		rwtype = rw::Light::AMBIENT;
		break;
	}
	if(rwtype == rw::Light::SPOT && state.fallsize-state.hotsize > 5.0f)
		rwtype = rw::Light::SOFTSPOT;
	if(state.ambientOnly)
		rwtype = rw::Light::AMBIENT;
	rw::Light *rwlight = rw::Light::create(rwtype);
	rwlight->setColor(state.color.r, state.color.g, state.color.b);
	if(rwtype >= rw::Light::POINT)
		rwlight->radius = state.attenEnd;
	if(state.type == SPOT_LGT)
		rwlight->setAngle(DegToRad(state.fallsize/2.0f));
	rwlight->setFrame(frame);
	clump->addLight(rwlight);
	//matrixMult(frame->matrix, flipZMat, frame->matrix);
}

void
convertCamera(rw::Frame *frame, rw::Clump *clump, INode *node)
{
	using namespace rw;
	GenCamera *gc = (GenCamera*)node->EvalWorldState(0).obj;
	CameraState state;
	Interval valid = FOREVER;
	gc->EvalCameraState(0, valid, &state);

	rw::Camera *rwcam = rw::Camera::create();
	rwcam->projection = !state.isOrtho;
	rwcam->nearPlane = state.hither;
	rwcam->farPlane = state.yon;
	rwcam->fogPlane = rwcam->nearPlane;
	rwcam->viewWindow.x = tan(state.fov/2.0f);
	rwcam->viewWindow.y = rwcam->viewWindow.x*3.0f/4.0f;
	rwcam->setFrame(frame);
	clump->addCamera(rwcam);
	//matrixMult(frame->matrix, flipZMat, frame->matrix);
}

void
convertAtomic(rw::Frame *frame, rw::Clump *clump, INode *node)
{
	using namespace rw;
	::Object *obj = node->EvalWorldState(0).obj;
	Geometry *geo = convertGeometry(node);
	Atomic *atomic = Atomic::create();
	atomic->geometry = geo;
	atomic->setFrame(frame);
	clump->addAtomic(atomic);
}

void
convertNode(rw::Clump *clump, rw::Frame *frame, INode *node, int ind = 0)
{
	using namespace rw;

	Matrix3 mat = node->GetNodeTM(0);
	if(!node->GetParentNode()->IsRootNode()){
		Matrix3 parentinv = Inverse(node->GetParentTM(0));
		mat = mat * parentinv;
	}
	MRow *m = mat.GetAddr();
	frame->matrix[0]  = m[0][0];
	frame->matrix[1]  = m[0][1];
	frame->matrix[2]  = m[0][2];
	frame->matrix[4]  = m[1][0];
	frame->matrix[5]  = m[1][1];
	frame->matrix[6]  = m[1][2];
	frame->matrix[8]  = m[2][0];
	frame->matrix[9]  = m[2][1];
	frame->matrix[10] = m[2][2];
	frame->matrix[12] = m[3][0];
	frame->matrix[13] = m[3][1];
	frame->matrix[14] = m[3][2];

	TCHAR *name = (TCHAR*)node->GetName();

	::Object *obj = node->EvalWorldState(0).obj;
	lprintf(_T("Obj: %s"), name);
	if(obj && node->Renderable() && !node->IsNodeHidden()){
		switch(obj->SuperClassID()){
		case GEOMOBJECT_CLASS_ID:
			if(obj->IsRenderable())
				convertAtomic(frame, clump, node);
			lprintf(_T(" geometry\n"));
			break;
		case CAMERA_CLASS_ID:
			lprintf(_T(" camera\n"));
			convertCamera(frame, clump, node);
			break;
		case LIGHT_CLASS_ID:
			lprintf(_T(" light\n"));
			convertLight(frame, clump, node);
			break;
		case SHAPE_CLASS_ID:
			lprintf(_T(" shape\n"));
			break;
		case HELPER_CLASS_ID:
			lprintf(_T(" helper\n"));
			break;
		}
	}

	for(int i = 0; i < ind; i++) lprintf(_T(" "));
	char *rwname = gta::getNodeName(frame);
#ifdef _UNICODE
	char fuckmax[MAX_PATH];
	wcstombs(fuckmax, name, MAX_PATH);
	strncpy(rwname, fuckmax, 24);
#else
	strncpy(rwname, name, 24);
#endif
	rwname[23] = '\0';
	lprintf(_T("exporting frame: %s\n"), name);

	int numChildren = node->NumberOfChildren();
	for(int i = 0; i < numChildren; i++){
		Frame *childframe = Frame::create();
		INode *childnode = node->GetChildNode(i);
		frame->addChild(childframe);
		convertNode(clump, childframe, childnode, ind+1);
	}
}

Value*
exportTest_cf(Value **arg_list, int count)
{
	using namespace rw;

	if(!rwInitialized)
		initRW();

	Interface *ip = GetCOREInterface();
	//dumpScene(ip->GetRootNode(), 0);
	int numSelected = ip->GetSelNodeCount();
	INode *rootnode = NULL;
	for(int i = 0; i < numSelected; i++){
		INode *n = ip->GetSelNode(i);
		rootnode = getRootOf(n);
		lprintf(_T(": %s : %s\n"), n->GetName(), rootnode->GetName());
	}
	if(rootnode == NULL){
		lprintf(_T("nothing selected\n"));
		return &ok;
	}

	Clump *clump = Clump::create();
	Frame *rootFrame = Frame::create();
	clump->setFrame(rootFrame);
	convertNode(clump, rootFrame, rootnode);
	//dumpFrameHier(clump->getFrame());
	StreamFile stream;
	if(stream.open("C:\\Users\\aap\\Desktop\\tmp\\rwtest\\export.dff", "wb") == NULL){
		lprintf(_T("couldn't open file\n"));
		return &ok;
	}
	hAnimDoStream = 0;
	clump->streamWrite(&stream);
	hAnimDoStream = 1;
	stream.close();
	clump->destroy();
	return &ok;
}

//def_visible_primitive(exportTest, "exportTest");

void
initImporter(void)
{
	define_struct_global(_T("convertAxes"), _T("dffImp"), getConvertAxes, setConvertAxes);
	define_struct_global(_T("autoSmooth"), _T("dffImp"), getAutoSmooth, setAutoSmooth);
	define_struct_global(_T("prepend"), _T("dffImp"), getPrepend, setPrepend);

/*	AllocConsole();
	freopen("CONIN$", "r", stdin);
	freopen("CONOUT$", "w", stdout);
	freopen("CONOUT$", "w", stderr);*/
}
